// ***
// IRC plugin v2 - http://www.ganggarrison.com/forums/index.php?topic=34618.0
// With both server-sent and vanilla plugin modes
//
// Copyright © 2013 Andrea Faulds
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
// ***

// Constants
// Plain plugins are passed no parameters (missing parameters == 0)
// But server-sent plugins are passed path (string) and packetId
// Hence is_string(argument0) will show us if we're in server-sent mode
global.IRCServerSentMode = is_string(argument0);
global.IRCPath = argument0;
global.IRCPacketId = argument1;
global.IRCFont = font_add("Lucida Console", 8, 0, 0, 32, 127);

// Packet IDs
global.IRC_PING = 0;
global.IRC_SERVER_CHANNEL = 1;

// IRC server settings
global.IRCHost = 'irc.esper.net';
global.IRCPort = 6667;
global.IRCUser = 'GG2';
global.IRCRealName = 'Gang Garrison 2 IRC plugin';

// Host can specify IRC channel for server
if (global.IRCServerSentMode) {
    if (global.isHost) {
        ini_open("gg2.ini");
        global.IRCChan = ini_read_string("Plugins", "irc_server_sent_server_channel", "#gg2_ingame");
        ini_write_string("Plugins", "irc_server_sent_server_channel", global.IRCChan);
        ini_close();
    }
// Client-side users can specify which they want to join
} else {
    ini_open("gg2.ini");
    global.IRCChan = ini_read_string("Plugins", "irc_vanilla_plugin_join_channel", "#gg2");
    ini_write_string("Plugins", "irc_vanilla_plugin_join_channel", global.IRCChan);
    ini_close();
}

// Variables

// Scripts
global.IRCFixNick = "
// string global.IRCFixNick(string nick)
// Makes sanitised (IRC-safe) nick.
// Return value string - sanitised nick
// string nick - input nick to be sanitised
var nick;
nick = argument0;

var safeNick, i, c;
safeNick = '';
for (i = 1; i <= min(string_length(nick), 30); i += 1) {
    c = string_char_at(nick, i);
    // [a-zA-Z0-9]
    if (('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || ('0' <= c && c <= '9')) {
        // IRC has no problems with these characters, append to IRC nickname
        safeNick += c;
    } else {
        // IRC might have problems with this character, append an underscore instead
        safeNick += '_';
    }
}

return safeNick;
";

global.IRCSetGG2Nick = "
// void global.IRCSetGG2Nick(string nick)
// Sets GG2 nick
// string nick - input nick to set GG2 nick to

var nick;
nick = argument0;

// Update our GG2 name to this nick
global.playerName = nick;
gg2_write_ini('Settings', 'PlayerName', nick);
write_ubyte(global.serverSocket, PLAYER_CHANGENAME);
write_ubyte(global.serverSocket, string_length(nick));
write_string(global.serverSocket, nick);
socket_send(global.serverSocket);
";

global.IRCSetIRCNick = "
// void global.IRCSetIRCNick(string nick)
// Sets IRC nick
// string nick - input nick to set IRC nick to

var nick;
nick = argument0;

with (global.IRCClient) {
    write_string(IRCSocket, 'NICK ' + nick + CRLF);
    socket_send(IRCSocket);
    attemptedNick = nick;
}
";

global.IRCPrePadString = "
// prePadString(string, num, char)
var string, num, char;
string = argument0;
num = argument1;
char = argument2;

while (string_length(string) < num) {
    string = char + string;
}

return string;
"

// Objects
global.IRCClient = object_add();
object_set_persistent(global.IRCClient, 1);
object_event_add(global.IRCClient, ev_create, 0, '
    mode = 0;   // 0 - Waiting for details
                // 1 - Connecting/connected
    IRCSocket = -1;
    IRCLog = ds_list_create();
    CRLF = chr(13) + chr(10);
    buf = "";
    retries = 0;
    ready = false;
    attemptedNick = "";


    // Server-sent
    if (global.IRCServerSentMode) {
        // Client - we need to ask server for details
        if (!global.isHost) {
            // Tell host server we exist
            var packet;
            packet = buffer_create();
            write_ubyte(packet, global.IRC_PING);
            PluginPacketSend(global.IRCPacketId, packet);
            buffer_destroy(packet);
        // Server - we can just join straight away
        } else {
            event_user(0);
        }
    // Client-side - we can also just join straight away
    } else {
        event_user(0);
    }
');
object_event_add(global.IRCClient, ev_other, ev_user0, '
    IRCSocket = tcp_connect(global.IRCHost, global.IRCPort);

    var now;
    now = date_current_datetime();
    startDate = string(date_get_year(now))
                + "-" + execute_string(global.IRCPrePadString, string(date_get_month(now)), 2, "0")
                + "-" + execute_string(global.IRCPrePadString, string(date_get_day(now)), 2, "0")
                + "T" + execute_string(global.IRCPrePadString, string(date_get_hour(now)), 2, "0")
                + "-" + execute_string(global.IRCPrePadString, string(date_get_minute(now)), 2, "0")
                + "-" + execute_string(global.IRCPrePadString, string(date_get_second(now)), 2, "0")
                + "Z";

    var newNick;
    newNick = execute_string(global.IRCFixNick, global.playerName);
    if (newNick != global.playerName)
        execute_string(global.IRCSetGG2Nick, global.playerName);
    execute_string(global.IRCSetIRCNick, global.playerName);
    write_string(IRCSocket, "USER " + global.IRCUser + " 0 * :" + global.IRCRealName + CRLF);
    socket_send(IRCSocket);
    mode = 1;
');
object_event_add(global.IRCClient, ev_step, ev_step_begin, '
    if (mode == 0) {
        // server-sent plugin, client, need to wait for details
        if (!global.isHost and global.IRCServerSentMode) {
            var buffer;
            buffer = PluginPacketGetBuffer(global.IRCPacketId);
            if (buffer != -1) {
                var type;
                type = read_ubyte(buffer);
                if (type == global.IRC_SERVER_CHANNEL) {
                    global.IRCChan = read_string(buffer, read_ubyte(buffer));
                    event_user(0);
                } else {
                    show_error("Unknown IRC plugin server-sent packet type: " + real(type), true);
                }
                PluginPacketPop(global.IRCPacketId);
            }
        }
        exit;
    }

    // server-sent plugin, host, need to service details requests
    if (global.isHost and global.IRCServerSentMode) {
        var buffer, player;
        while (PluginPacketGetBuffer(global.IRCPacketId) != -1) {
            buffer = PluginPacketGetBuffer(global.IRCPacketId);
            player = PluginPacketGetPlayer(global.IRCPacketId);
            
            var type;
            type = read_ubyte(buffer);
            
            // Ignore bad packets
            if (type == global.IRC_PING) {
                var response;
                response = buffer_create();
                write_ubyte(response, global.IRC_SERVER_CHANNEL);
                write_ubyte(response, string_length(global.IRCChan));
                write_string(response, global.IRCChan);
                PluginPacketSendTo(global.IRCPacketId, response, player);
                buffer_destroy(response);
            }
            
            PluginPacketPop(global.IRCPacketId);
        }
    }
    
    if (socket_has_error(IRCSocket)) {
        show_message("Socket error: " + socket_error(IRCSocket));
        game_end();
        exit;
    }
    var available;
    available = tcp_receive_available(IRCSocket);

    buf += read_string(IRCSocket, available);
    
    var lines, i;
    lines = split(buf, CRLF);
    for (i = 0; i < ds_list_size(lines); i += 1) {
        if (i == ds_list_size(lines) - 1) {
            buf = ds_list_find_value(lines, i);
        } else {
            var parts;
            parts = split(ds_list_find_value(lines, i), " ");
            var prefix;
            prefix = "";
            // Remove prefix
            if (string_char_at(ds_list_find_value(parts, 0), 1) == ":") {
                prefix = ds_list_find_value(parts, 0);
                ds_list_delete(parts, 0);
            }
            var type;
            type = ds_list_find_value(parts, 0);
            switch (type) {
            case "PING":
                var token;
                token = ds_list_find_value(parts, 1);
                if (string_char_at(token, 1) == ":") {
                    token = string_copy(token, 2, string_length(token) - 1);
                }
                write_string(IRCSocket, "PONG " + token + CRLF);
                write_string(IRCSocket, "PONG " + ds_list_find_value(parts, 1) + CRLF);
                socket_send(IRCSocket);
            break;
            // ERR_NICKNAMEINUSE
            case "433":
                var newNick, len;
                len = min(string_length(attemptedNick), 20 - string_length(string(retries)));
                newNick = string_copy(attemptedNick, 1, len) + string(retries);
                retries += 1;
                execute_string(global.IRCSetIRCNick, newNick);
                if (!ready)
                    execute_string(global.IRCSetGG2Nick, newNick);
            break;
            case "NOTICE":
            // RPL_MOTD
            case "372":
                var line;
                line = "* ";
                var j;
                for (j = 2; j < ds_list_size(parts); j += 1) {
                    var part;
                    part = ds_list_find_value(parts, j);
                    if (string_char_at(part, 1) == ":" and j == 2) {
                        line += string_copy(part, 2, string_length(part) - 1) + " ";
                    } else {
                        line += part + " ";
                    }
                }
                ds_list_add(IRCLog, line);
            break;
            // RPL_MOTDSTART
            case "375":
                ds_list_add(IRCLog, "<MOTD>");
            break;
            // RPL_ENDOFMOTD
            case "376":
                ds_list_add(IRCLog, "</MOTD>");
                ready = true;
                write_string(IRCSocket, "JOIN " + global.IRCChan + CRLF);
                socket_send(IRCSocket);
            break;
            // RPL_TOPIC
            case "332":
                var line;
                line = "";
                var j;
                for (j = 3; j < ds_list_size(parts); j += 1) {
                    var part;
                    part = ds_list_find_value(parts, j);
                    if (string_char_at(part, 1) == ":" and j == 3) {
                        line += string_copy(part, 2, string_length(part) - 1) + " ";
                    } else {
                        line += part + " ";
                    }
                }
                ds_list_add(IRCLog, "* Topic for " + ds_list_find_value(parts, 2) + ": " + line);
            break;
            // Seems to be "set by" for topic
            case "333":
                ds_list_add(IRCLog, "* Topic for " + ds_list_find_value(parts, 2) + " set by " + ds_list_find_value(parts, 3));
            break;
            case "JOIN":
                ds_list_add(IRCLog, "* " + string_copy(prefix, 2, string_length(prefix) - 1) + " joined " + ds_list_find_value(parts, 1));
            break;
            case "MODE":
                var modes;
                // Channel modes
                if (ds_list_find_value(parts, 1) == global.IRCChan) {
                    prefix = string_copy(prefix, 2, string_length(prefix) - 1);
                    var nickBits;
                    nickBits = split(prefix, "!");
                    var line;
                    line = "";
                    var j;
                    for (j = 2; j < ds_list_size(parts); j += 1) {
                        var part;
                        part = ds_list_find_value(parts, j);
                        if (string_char_at(part, 1) == ":" and j == 2) {
                            line += string_copy(part, 2, string_length(part) - 1) + " ";
                        } else {
                            line += part + " ";
                        }
                    }
                    ds_list_add(IRCLog, "* " + ds_list_find_value(nickBits, 0) + " sets mode: " + line);
                    ds_list_destroy(nickBits);
                } else {
                    modes = ds_list_find_value(parts, 2);
                    string_copy(modes, 2, string_length(modes) - 1);
                    ds_list_add(IRCLog, "* Your mode was set to mode " + modes);
                }
            break;
            case "PRIVMSG":
                prefix = string_copy(prefix, 2, string_length(prefix) - 1);
                var nickBits;
                nickBits = split(prefix, "!");
                var line;
                line = "";
                var j;
                for (j = 2; j < ds_list_size(parts); j += 1) {
                    var part;
                    part = ds_list_find_value(parts, j);
                    if (string_char_at(part, 1) == ":" and j == 2) {
                        line += string_copy(part, 2, string_length(part) - 1) + " ";
                    } else {
                        line += part + " ";
                    }
                }
                // /me :D
                if (string_copy(line, 1, 8) == chr(1) + "ACTION ") {
                    line = string_copy(line, 9, string_length(line) - 8);
                    // Channel or private message?
                    if (ds_list_find_value(parts, 1) == global.IRCChan) {
                        ds_list_add(IRCLog, "!* " + ds_list_find_value(nickBits, 0) + " " +line);
                    } else {
                        ds_list_add(IRCLog, "(PRIVMSG) !* " + ds_list_find_value(nickBits, 0) + " " + line);
                    }
                } else {
                    // Channel or private message?
                    if (ds_list_find_value(parts, 1) == global.IRCChan) {
                        ds_list_add(IRCLog, "<" + ds_list_find_value(nickBits, 0) + "> " + line);
                    } else {
                        ds_list_add(IRCLog, "(PRIVMSG) <" + ds_list_find_value(nickBits, 0) + "> " + line);
                    }
                }
                ds_list_destroy(nickBits);
            break;
            case "QUIT":
                prefix = string_copy(prefix, 2, string_length(prefix) - 1);
                var nickBits;
                nickBits = split(prefix, "!");
                var line;
                line = "";
                var j;
                for (j = 1; j < ds_list_size(parts); j += 1) {
                    var part;
                    part = ds_list_find_value(parts, j);
                    if (string_char_at(part, 1) == ":" and j == 1) {
                        line += string_copy(part, 2, string_length(part) - 1) + " ";
                    } else {
                        line += part + " ";
                    }
                }
                ds_list_add(IRCLog, "* " + ds_list_find_value(nickBits, 0) + " quit (" + line + ")");
                ds_list_destroy(nickBits);
            break;
            case "PART":
                prefix = string_copy(prefix, 2, string_length(prefix) - 1);
                var nickBits;
                nickBits = split(prefix, "!");
                var line;
                line = "";
                var j;
                for (j = 2; j < ds_list_size(parts); j += 1) {
                    var part;
                    part = ds_list_find_value(parts, j);
                    if (string_char_at(part, 1) == ":" and j == 2) {
                        line += string_copy(part, 2, string_length(part) - 1) + " ";
                    } else {
                        line += part + " ";
                    }
                }
                ds_list_add(IRCLog, "* " + ds_list_find_value(nickBits, 0) + " left " + ds_list_find_value(parts, 1) + " (" + line + ")");
                ds_list_destroy(nickBits);
            break;
            case "NICK":
                prefix = string_copy(prefix, 2, string_length(prefix) - 1);
                var nickBits;
                nickBits = split(prefix, "!");
                var newNick;
                newNick = ds_list_find_value(parts, 1);
                newNick = string_copy(newNick, 2, string_length(newNick) - 1);
                ds_list_add(IRCLog, "* " + ds_list_find_value(nickBits, 0) + " is now known as " + newNick);
                // It is my name that is being changed!
                if (string_lower(ds_list_find_value(nickBits, 0)) == string_lower(global.playerName))
                    execute_string(global.IRCSetGG2Nick, newNick);
                ds_list_destroy(nickBits);
            break;
            // RPL_NAMREPLY
            case "353":
            // RPL_ENDOFNAMES
            case "366":
                // Skip
            break;
            default:
                ds_list_add(IRCLog, "? " + ds_list_find_value(lines, i));
            break;
            }
            ds_list_destroy(parts);
        }
    }
    ds_list_destroy(lines);
');
object_event_add(global.IRCClient, ev_other, ev_game_end, '
    if (!directory_exists(working_directory + "\IRCLogs")) {
        directory_create(working_directory + "\IRCLogs");
    }
    var fp, i;
    fp = file_text_open_write(working_directory + "\IRCLogs\" + startDate + ".txt");
    for (i = 0; i < ds_list_size(IRCLog); i += 1) {
        file_text_write_string(fp, ds_list_find_value(IRCLog, i));
        file_text_writeln(fp);
    }
    file_text_close(fp);
');

global.IRCWindow = object_add();
object_set_depth(global.IRCWindow, -110002);
object_event_add(global.IRCWindow, ev_create, 0, '
    open = false;
    showback = true;
');
object_event_add(global.IRCWindow, ev_step, ev_step_end, '
    if ((!instance_exists(InGameMenuController) && !instance_exists(OptionsController) && !instance_exists(ControlsController) && !instance_exists(HUDOptionsController)) || instance_exists(global.cheatycheat)) {
        // "open" to enter text
        if (!open && keyboard_check_pressed(ord("Y"))) {
            open = true;
            keyboard_string = "";
            instance_create(0, 0, global.cheatycheat);
        // toggle background
        } else if (!open && keyboard_check_pressed(ord("T"))) {
            showback = !showback;
        // Send message and "close"
        } else if (open && keyboard_check_pressed(vk_enter)) {
            with (global.cheatycheat) {
                instance_destroy();
            }
            open = false;
            // Command
            if (string_char_at(keyboard_string, 1) == "/") {
                var commandLine, commandParts, command, arguments;
                commandLine = string_copy(keyboard_string, 2, string_length(keyboard_string) - 1);
                commandParts = split(commandLine, " ", 1);
                command = ds_list_find_value(commandParts, 0);
                arguments = ds_list_find_value(commandParts, 1);
                if (!is_string(command))
                    command = "";
                if (!is_string(arguments))
                    arguments = "";
                if (command == "msg") {
                    var argumentParts, nick, msg;
                    argumentParts = split(arguments, " ", 1);
                    if (ds_list_size(argumentParts) < 2) {
                        with (global.IRCClient)
                            ds_list_add(IRCLog, "Error: /msg takes at least 2 parameters");
                    } else {
                        nick = ds_list_find_value(argumentParts, 0);
                        msg = ds_list_find_value(argumentParts, 1);
                        with (global.IRCClient) {
                            write_string(IRCSocket, "PRIVMSG " + nick + " :" + msg + CRLF);
                            socket_send(IRCSocket);
                            ds_list_add(IRCLog, "(PRIVMSG->" + nick + ") <" + global.playerName + "> " + msg);
                        }
                    }
                } else if (command == "nick") {
                    var newNick;
                    newNick = execute_string(global.IRCFixNick, arguments);
                    execute_string(global.IRCSetIRCNick, newNick);
                } else if (command == "me") {
                    with (global.IRCClient) {
                        write_string(IRCSocket, "PRIVMSG " + global.IRCChan + " :" + chr(1) + "ACTION " + arguments + CRLF);
                        socket_send(IRCSocket);
                        ds_list_add(IRCLog, "!* " + global.playerName + " " + arguments);
                    }
                } else if (command == "help") {
                    with (global.IRCClient) {
                        if (arguments == "msg")
                            ds_list_add(IRCLog, "/msg <nick> <message> - sends a private message to someone, e.g. /msg ajf Needs more Pinkie Pie.");
                        else if (arguments == "me")
                            ds_list_add(IRCLog, "/me <action> - makes it look like you did an action, e.g. /me kills a spy");
                        else if (arguments == "nick")
                            ds_list_add(IRCLog, "/nick <nick> - changes your IRC and GG2 nickname, e.g. /nick Player");
                        else
                            ds_list_add(IRCLog, "Available Commands: msg, me, nick (do /help commandname for more info)");
                    }
                } else {
                    with (global.IRCClient)
                        ds_list_add(IRCLog, "Error: Unrecognised command");
                }
            } else if (string_length(keyboard_string) != 0) {
                with (global.IRCClient) {
                    write_string(IRCSocket, "PRIVMSG " + global.IRCChan + " :" + keyboard_string + CRLF);
                    socket_send(IRCSocket);
                    ds_list_add(IRCLog, "<" + global.playerName + "> " + keyboard_string);
                }
            }
            keyboard_string = "";
        }
    }
');
object_event_add(global.IRCWindow, ev_draw, 0, '
    var xoffset, yoffset, xsize, ysize, xmax, ymax;
    xoffset = view_xview[0];
    yoffset = view_yview[0];
    xsize = view_wview[0];
    ysize = view_hview[0];
    xmax = 380;
    ymax = 180;
    x = view_xview[0]+7;
    y = view_yview[0]+ysize-80;

    if (showback or open) {
        draw_set_alpha(0.2928932188134525);
        draw_set_color(c_black);
        draw_rectangle(x, y, x + xmax, y - ymax, false);
        draw_set_alpha(0.2928932188134525 * 0.2928932188134525);
        draw_set_color(c_black);
        draw_rectangle(x, y, x + xmax, y - ymax, true);
    }
    
    draw_set_halign(fa_left);
    draw_set_valign(fa_bottom);
    draw_set_font(global.IRCFont);
    draw_set_alpha(1);
    var _y, linesDrawn, i, line;
    _y = 0;
    linesDrawn = 0;
    // Text entry
    if (open) {
        draw_set_color(c_orange);
        line = "> " + string_replace(keyboard_string, "#", "\#");
        draw_text(x, y, line);
        _y += string_height(line);
    }
    draw_set_color(c_white);
    for (i = ds_list_size(global.IRCClient.IRCLog) - 1; i >= 0; i -= 1) {
        line = string_replace(ds_list_find_value(global.IRCClient.IRCLog, i), "#", "\#");
        if (_y + string_height_ext(line, -1, xmax) > ymax) {
            break;
        }
        if (!showback and linesDrawn >= 2) {
            break;
        }
        // Dark green for info lines
        if (string_char_at(line, 1) == "*") {
            draw_set_color(c_lime);
        // /me lines
        } else if (string_char_at(line, 1) == "!") {
            draw_set_color(c_fuchsia);
        // ? lines
        } else if (string_char_at(line, 1) == "?") {
            draw_set_color(c_red);
        } else {
            draw_set_color(c_white);
        }
        draw_text_ext(x, y - _y, line, -1, xmax);
        _y += string_height_ext(line, -1, xmax);
        linesDrawn += 1;
    }
    draw_set_font(global.gg2Font);
');

// Add events to Player to keep nickname safe
object_event_add(Player, ev_step, ev_step_begin, "
    if (!variable_local_exists('oldName')) {
        oldName = '';
    }
    if (global.myself == id and name != oldName) {
        oldName = name;
        
        var safeNick;
        safeNick = execute_string(global.IRCFixNick, name);
        if (safeNick != name)
            execute_string(global.IRCSetIRCNick, safeNick);
    }
");

// I STOLE THIS FROM LORGAN'S CHAT CODE SO GIVE HIM THE CREDIT
//create a cheaty cheat object to prevent you from moving and opening chat menus while typing :p
global.cheatycheat = object_add();
object_set_parent(global.cheatycheat,InGameMenuController);
object_event_add(global.cheatycheat,ev_step,ev_step_normal,'//no');
object_event_add(global.cheatycheat,ev_draw,0,'//no');
object_event_add(global.cheatycheat,ev_keypress,vk_enter,'//no');
object_event_add(global.cheatycheat,ev_keypress,vk_escape,'
    with(global.IRCWindow) open = false;
    instance_create(0,0,InGameMenuController);
    instance_destroy();
');

// Add event to PlayerControl to create client/window
object_event_add(PlayerControl, ev_step, ev_step_end, '
    if (!instance_exists(global.IRCClient)) {
        instance_create(0, 0, global.IRCClient);
    }
    if (!instance_exists(global.IRCWindow)) {
        instance_create(0, 0, global.IRCWindow);
    }
');
